"""Process the raw MSP-PODCAST dataset.

This assumes the file structure from the original compressed file:
/.../
    Labels/
        labels_concensus.csv
        ...
    Audios/
        *.wav
    readme.txt
    ...
"""

import json
from pathlib import Path

import click
import numpy as np
import pandas as pd

from ertk.dataset import write_annotations, write_filelist
from ertk.stats import alpha

emotion_map = {
    "A": "anger",
    "H": "happiness",
    "S": "sadness",
    "N": "neutral",
    "U": "surprise",
    "F": "fear",
    "D": "disgust",
    "C": "contempt",
    "O": "other",
    "X": "unknown",
}

rating_map = {"angry": "anger", "happy": "happiness", "sad": "sadness"}


def _filter_other(s: str) -> str:
    if s.startswith("Other"):
        s = "Other"
    s = s.lower()
    return rating_map.get(s, s)


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option(
    "--resample/--noresample", default=False, help="Resample audio to local directory."
)
def main(input_dir: Path, resample: bool):
    """Process the MSP-IMPROV dataset at location INPUT_DIR."""

    labels_dir = input_dir / "Labels"
    labels_concensus = pd.read_csv(
        labels_dir / "labels_concensus.csv",
        header=0,
        low_memory=False,
        na_values=["Unknown"],
        dtype={"SpkrID": str},
    )
    labels_concensus.set_index("FileName", inplace=True)
    labels_concensus.index = labels_concensus.index.map(lambda x: x[:-4])
    for dim in ["EmoAct", "EmoVal", "EmoDom"]:
        write_annotations(labels_concensus[dim].to_dict(), dim)
    labels = labels_concensus["EmoClass"].to_dict()

    audios_dir = input_dir / "Audios"
    paths = list(audios_dir.glob("*.wav"))
    write_filelist(paths, "files_all")
    split_sets = labels_concensus.groupby("Split_Set").groups
    for fileset, names in split_sets.items():
        write_filelist(
            names.map(lambda x: audios_dir / (x + ".wav")), f"files_{fileset}"
        )

    write_annotations({n: emotion_map[labels[n]] for n in labels}, "label")
    write_annotations(labels_concensus["Gender"].to_dict(), "gender")
    write_annotations(labels_concensus["SpkrID"].to_dict(), "speaker")
    write_annotations(labels_concensus["Split_Set"].to_dict(), "split")
    write_annotations({p.stem: "en" for p in paths}, "language")

    #
    # Ratings analysis
    #
    with open(labels_dir / "labels_detailed.json") as fid:
        labels_detailed = json.load(fid)
        _ratings = []
        for name, workers in labels_detailed.items():
            for worker, annotations in workers.items():
                _ratings.append(
                    (
                        name[:-4],
                        worker,
                        _filter_other(annotations["EmoClass_Major"]),
                        annotations["EmoAct"],
                        annotations["EmoDom"],
                        annotations["EmoVal"],
                    )
                )
    ratings = pd.DataFrame(
        sorted(_ratings), columns=["name", "rater", "label", "act", "val", "dom"]
    )
    ratings.set_index(["name", "rater"]).to_csv("ratings.csv")

    num_ratings = ratings.groupby("name").size().to_frame("total")
    label_count = ratings.groupby(["name", "label"]).size().to_frame("freq")
    # Count of majority label per utterance
    mode_count = (
        label_count.reset_index()
        .sort_values("freq", ascending=False)
        .drop_duplicates(subset="name")
        .set_index("name")
        .join(num_ratings)
        .sort_index()
    )

    # Include only names with a label which is strictly a plurality
    mode_count = mode_count[
        ratings.groupby("name")["label"]
        .agg(lambda x: "".join(x.mode()))
        .map(lambda x: len(x) == 1)
        .sort_index()
    ]

    # Agreement is mean proportion of labels which are plurality label
    agreement = np.mean(mode_count["freq"] / mode_count["total"])
    print(f"Mean label agreement: {agreement:.3f}")

    clips = ratings.join(mode_count["label"], "name", rsuffix="_vote", how="inner")
    accuracy = (clips["label"] == clips["label_vote"]).sum() / len(clips)
    print(f"Human accuracy: {accuracy:.3f}")

    # Simple way to get int matrix of labels for raters x clips
    data = (
        ratings.set_index(["rater", "name"])["label"]
        .astype("category")
        .cat.codes.unstack()
        + 1
    )
    data[data.isna()] = 0
    data = data.astype(int).to_numpy()
    print(f"Krippendorf's alpha: {alpha(data):.3f}")


if __name__ == "__main__":
    main()



import logging
from pathlib import Path
from typing import Tuple

import click
from omegaconf.errors import OmegaConfBaseException

from ertk.cli._utils import dataset_args, debug_args
from ertk.config import get_arg_mapping
from ertk.dataset import DataLoadConfig, load_datasets_config, load_multiple
from ertk.train import ExperimentConfig


def write_list(dataset, output_list):
    with open(output_list, "w") as fid:
        fid.write("\n".join(dataset.get_audio_paths()))


@click.command()
@dataset_args
@debug_args
@click.option("--output_list", type=Path)
def main(
    corpus_info: Tuple[Path],
    data_config: Path,
    subset: Tuple[str],
    sel_groups: Tuple[str],
    remove_groups: Tuple[str],
    map_groups: Tuple[str],
    clip_seq: int,
    pad_seq: int,
    verbose: int,
    output_list: Path,
):
    """Print info about a dataset or combination of datasets."""
    if verbose > 0:
        logging.basicConfig(level=logging.DEBUG if verbose > 1 else logging.INFO)

    if data_config:
        try:
            config = DataLoadConfig.from_file(data_config)
        except OmegaConfBaseException:
            config = ExperimentConfig.from_file(data_config).data
        config.features = None
        for c in config.datasets:
            config.datasets[c].features = None
        dataset = load_datasets_config(config)
        print(dataset)
        if output_list:
            write_list(dataset, output_list)
        return

    if len(corpus_info) == 0:
        raise click.UsageError("No corpus info files specified")

    if len(subset) == 1 and ":" not in subset[0]:
        dataset = load_multiple(corpus_info, subsets=subset[0], label="")
    else:
        subset_map = {}
        for m in subset:
            subset_map.update(get_arg_mapping(m))
        dataset = load_multiple(corpus_info, subsets=subset_map, label="")

    grp_map = {}
    for m in map_groups:
        grp_map.update(get_arg_mapping(m))
    grp_sel = {}
    for m in sel_groups:
        grp_sel.update(get_arg_mapping(m))
    grp_del = {}
    for m in remove_groups:
        grp_del.update(get_arg_mapping(m))
    dataset.map_and_select(grp_map, grp_sel, grp_del)

    if clip_seq:
        dataset.clip_arrays(clip_seq)
    if pad_seq:
        dataset.pad_arrays(pad_seq)

    print(dataset)
    if output_list:
        write_list(dataset, output_list)


if __name__ == "__main__":
    main()



"""Process the raw CaFE dataset.

This assumes the file structure from the original compressed file:
/.../
    Colère/
        Faible/
            *.aiff [for 192k data]
            *.wav  [for 48k data]
        Fort/
    ...
"""

from pathlib import Path

import click

from ertk.dataset import resample_audio, write_annotations, write_filelist

emotion_map = {
    "C": "anger",
    "D": "disgust",
    "J": "happiness",
    "N": "neutral",
    "P": "fear",
    "S": "surprise",
    "T": "sadness",
}

sentence_map = {
    "1": "Un cheval fou dans mon jardin",
    "2": "Deux ânes aigris au pelage brun",
    "3": "Trois cygnes aveugles au bord du lac",
    "4": "Quatre vieilles truies éléphantesques",
    "5": "Cinq pumas fiers et passionnés",
    "6": "Six ours aimants domestiqués",
}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the CaFE dataset at location INPUT_DIR and resample audio
    to 16 kHz 16-bit WAV audio.
    """

    paths = list(input_dir.glob("**/*.wav"))
    if len(paths) == 0:
        paths = list(input_dir.glob("**/*.aiff"))
    if resample:
        resample_dir = Path("resampled")
        resample_audio(paths, resample_dir)
        write_filelist(resample_dir.glob("*.wav"), "files_all")

    write_annotations({p.stem: emotion_map[p.stem[3]] for p in paths}, "label")
    speaker_dict = {p.stem: p.stem[:2] for p in paths}
    write_annotations(speaker_dict, "speaker")
    write_annotations(
        {k: ["F", "M"][int(v) % 2] for k, v in speaker_dict.items()},
        "gender",
    )
    sentence_dict = {p.stem: p.stem[-1] for p in paths}
    write_annotations(sentence_dict, "sentence")
    transcript = {k: sentence_map[v] for k, v in sentence_dict.items()}
    write_annotations(transcript, "transcript")
    write_annotations({p.stem: "fr" for p in paths}, "language")
    write_annotations({p.stem: "ca" for p in paths}, "country")


if __name__ == "__main__":
    main()



"""
Audioset feature extractors
===========================

This module contains feature extractors for the Audioset dataset.

.. autosummary::
    :toctree:

    VGGishExtractor
    VGGishExtractorConfig
    YamnetExtractor
    YamnetExtractorConfig
"""

from .audioset import (
    VGGishExtractor,
    VGGishExtractorConfig,
    YamnetExtractor,
    YamnetExtractorConfig,
)



"""
Scikit-learn backend
====================

This module contains the scikit-learn backend for ERTK. It is used to
provide a consistent interface for scikit-learn models, and to provide
additional functionality.

.. autosummary::
    :toctree: generated/

    models
    classification
    utils
"""

from .models import get_sk_model, get_sk_model_fn



"""Processing using SpeechBrain models."""

from dataclasses import dataclass
from enum import Enum
from typing import List, Optional

import numpy as np
import torch
from omegaconf import MISSING

from ertk.config import ERTKConfig
from ertk.utils import is_mono_audio

from ._base import AudioClipProcessor, FeatureExtractor

__all__ = ["SpeechBrainExtractorConfig", "SpeechBrainExtractor"]


class Task(Enum):
    """Task to perform."""

    WHISPER = "whisper"
    """Whisper speech recognition. Use for Whisper models."""
    EMBEDDINGS = "embeddings"
    """Extract embeddings."""
    ASR = "asr"
    """Automatic speech recognition."""


class Agg(Enum):
    """Aggregation method for extracting embeddings."""

    MEAN = "mean"
    MAX = "max"
    NONE = "none"


@dataclass
class SpeechBrainExtractorConfig(ERTKConfig):
    """SpeechBrain feature extractor configuration."""

    model: str = MISSING
    """Model name or path."""
    device: str = "cuda"
    """Device to run model on."""
    task: Task = Task.EMBEDDINGS
    """Task to perform."""
    agg: Optional[Agg] = Agg.MEAN
    """Aggregation method for embeddings."""
    max_input_len: int = 1500000
    """Maximum input length."""
    normalise: bool = True
    """Normalise audio."""


class SpeechBrainExtractor(
    FeatureExtractor,
    AudioClipProcessor,
    fname="speechbrain",
    config=SpeechBrainExtractorConfig,
):
    """Processing using SpeechBrain models."""

    config: SpeechBrainExtractorConfig

    def __init__(self, config: SpeechBrainExtractorConfig) -> None:
        from speechbrain.pretrained import (
            EncoderClassifier,
            EncoderDecoderASR,
            WhisperASR,
        )

        super().__init__(config)

        print(f"Loading model from {config.model}")
        if config.task == Task.EMBEDDINGS:
            self.model = EncoderClassifier.from_hparams(
                config.model, run_opts={"device": config.device}
            )
        elif config.task == Task.WHISPER:
            self.model = WhisperASR.from_hparams(
                config.model, run_opts={"device": config.device}
            )
        elif config.task == Task.ASR:
            self.model = EncoderDecoderASR.from_hparams(
                config.model, run_opts={"device": config.device}
            )

        self.model.to(config.device).eval()
        torch.set_grad_enabled(False)
        if config.task == Task.EMBEDDINGS:
            wav = torch.zeros(1, 16000, device=config.device)
            self._dim = self.model.encode_batch(wav).shape[2]

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        x = np.squeeze(x)
        if not is_mono_audio(x):
            raise ValueError("Audio must be mono")
        wav = torch.from_numpy(x).to(self.config.device)
        wav = wav[: self.config.max_input_len].unsqueeze(0)
        assert kwargs.pop("sr") == 16000
        if self.config.task in [Task.WHISPER, Task.ASR]:
            text, _ = self.model.transcribe_batch(wav, torch.tensor([1.0]))
            return np.array([" ".join(text[0])], dtype=object)
        # Embeddings
        output = self.model.encode_batch(wav, normalize=self.config.normalise)
        output = output.squeeze(0)
        if self.config.agg == Agg.MEAN:
            output = output.mean(0)
        elif self.config.agg == Agg.MAX:
            output = output.max(0)
        return output.cpu().numpy()

    @property
    def dim(self) -> int:
        if self.config.task in [Task.WHISPER, Task.ASR]:
            return 1
        return self._dim

    @property
    def is_sequence(self) -> bool:
        # Single text string
        return False

    @property
    def feature_names(self) -> List[str]:
        if self.config.task in [Task.WHISPER, Task.ASR]:
            return ["text"]
        return [f"{self.config.model}_{i}" for i in range(self.dim)]



from pathlib import Path

import pytest

from ertk.cli.dataset.remove_instances import main


def test_noargs():
    with pytest.raises(SystemExit) as excinfo:
        main([])
    assert excinfo.value.code == 2


def test_badargs():
    with pytest.raises(SystemExit) as excinfo:
        main(["x", "y"])
    assert excinfo.value.code == 2


test_data_dir = Path(__file__).parent / "../../test_data"


def test_remove_instances(tmp_path):
    with pytest.raises(SystemExit) as excinfo:
        main(
            [
                f"{test_data_dir}/features/features_2d.csv",
                "--names",
                f"{test_data_dir}/subset.txt",
                f"{tmp_path}/removed.csv",
            ]
        )
    assert excinfo.value.code == 0



"""Implementation of the model architecture from [1]_.

References
----------
.. [1] Z. Aldeneh and E. Mower Provost, 'Using regional saliency for
       speech emotion recognition', in 2017 IEEE International
       Conference on Acoustics, Speech and Signal Processing (ICASSP),
       Mar. 2017, pp. 2741-2745, doi: 10.1109/ICASSP.2017.7952655.
"""

from dataclasses import dataclass, field
from typing import List

import torch
import torch.nn.functional as F
from torch import nn

from ._base import PyTorchModelConfig, SimpleClassificationModel

__all__ = ["Aldeneh2017Config", "Aldeneh2017Model"]


@dataclass
class Aldeneh2017Config(PyTorchModelConfig):
    conv_dims: List[int] = field(default_factory=lambda: [384] * 4)
    kernel_sizes: List[int] = field(default_factory=lambda: [8, 16, 32, 64])
    dense_dims: List[int] = field(default_factory=lambda: [1024, 1024])


class Aldeneh2017Model(
    SimpleClassificationModel, fname="aldeneh2017", config=Aldeneh2017Config
):
    def __init__(self, config: Aldeneh2017Config):
        super().__init__(config)

        if len(config.conv_dims) != len(config.kernel_sizes):
            raise ValueError("`conv_dims` must be the same size as `kernel_sizes`.")

        self.convs = nn.ModuleList()
        for dim, ksize in zip(config.conv_dims, config.kernel_sizes):
            self.convs.append(nn.Conv1d(config.n_features, dim, kernel_size=ksize))

        dense_in = sum(config.conv_dims)
        dense_dims = [dense_in] + list(config.dense_dims)
        self.dense = nn.Sequential()
        for d1, d2 in zip(dense_dims[:-1], dense_dims[1:]):
            self.dense.append(nn.Sequential(nn.Linear(d1, d2), nn.ReLU(inplace=True)))
        self.dense.append(nn.Linear(dense_dims[-1], config.n_classes))

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # type: ignore
        # (B, T, D) -> (B, D, T)
        x = x.transpose(1, 2)
        cs = []
        for conv in self.convs:
            c, _ = torch.max(F.relu(conv(x)), 2)
            cs.append(c)
        x = torch.cat(cs, 1).squeeze(1)
        out = self.dense(x)
        return out



"""Process the raw MLEnd Spoken Numerals dataset.

This assumes the file structure from the original dataset:
/.../
    MLEndSND_Public/
        *.wav
    MLEndSND_Audio_Attributes.csv
    MLEndSND_Speakers_Demographics.csv
"""

from pathlib import Path

import click
import pandas as pd

from ertk.dataset import resample_audio, write_annotations, write_filelist


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the MLEnd dataset at location INPUT_DIR and resample audio
    to 16 kHz 16-bit WAV audio.
    """

    paths = list((input_dir / "MLEndSND_Public").glob("*.wav"))
    if resample:
        resample_dir = Path("resampled")
        resample_audio(paths, resample_dir)
        paths = list(resample_dir.glob("*.wav"))
    write_filelist(paths, "files_all")

    annots = pd.read_csv(input_dir / "MLEndSND_Audio_Attributes.csv", dtype=str)
    annots.set_index("Public filename", inplace=True)
    spk_info = pd.read_csv(input_dir / "MLEndSND_Speakers_Demographics.csv", dtype=str)
    spk_info.set_index("Speaker", inplace=True)
    df = annots.join(spk_info, "Speaker")

    write_annotations(df["Intonation"], "label")
    write_annotations(df["Speaker"], "speaker")
    write_annotations(df["Numeral"], "numeral")
    write_annotations(df["Nationality"], "nationality")
    write_annotations(df["Language1"], "l1")
    write_annotations({p.stem: "en" for p in paths}, "language")


if __name__ == "__main__":
    main()



"""Process the raw MELD dataset.

This assumes the following file structure from the original compressed files:
/.../
    dev_splits_complete/
        *.mp4
    train_splits/
        *.mp4
    output_repeated_splits_test/
        *.mp4
    *.csv
    ...
"""

from pathlib import Path

import click
import pandas as pd

from ertk.dataset import resample_rename_clips, write_annotations, write_filelist


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the MELD dataset at location INPUT_DIR and resample audio
    to 16 kHz 16-bit WAV audio.
    """
    splits = ["train", "dev", "test"]
    split_dirs = ["train_splits", "dev_splits_complete", "output_repeated_splits_test"]
    csvs = ["train_sent_emo.csv", "dev_sent_emo.csv", "test_sent_emo.csv"]

    resample_dir = Path("resampled")

    name_map = {}
    all_paths = []
    for split, x in zip(splits, split_dirs):
        paths = list(input_dir.glob(f"{x}/dia*.mp4"))
        all_paths.extend(paths)
        name_map.update({x: resample_dir / f"{split}_{x.stem}.wav" for x in paths})

    if resample:
        resample_rename_clips(mapping=name_map)

    dfs = []
    for split, x in zip(splits, csvs):
        df = pd.read_csv(input_dir / x)
        df.index = df.apply(
            lambda x: f"{split}_dia{x['Dialogue_ID']}_utt{x['Utterance_ID']}", axis=1
        )
        if split == "train":
            # Train dia125_utt3 is broken
            df = df.drop("train_dia125_utt3")
        df["split"] = split
        write_filelist({resample_dir / f"{x}.wav" for x in df.index}, f"files_{split}")
        dfs.append(df)
    df = pd.concat(dfs)
    df = df.drop(index={"dev_dia110_utt7"})
    write_filelist({resample_dir / f"{x}.wav" for x in df.index}, "files_all")

    write_annotations(df["Emotion"], "label")
    write_annotations(df["Speaker"], "speaker")
    write_annotations(df["split"], "split")
    write_annotations(df["Utterance"], "transcript")
    write_annotations({x: "us" for x in df.index}, "country")
    write_annotations({x: "en" for x in df.index}, "language")


if __name__ == "__main__":
    main()



"""Process the raw URDU dataset.

This assumes the file structure from the original compressed file:
/.../
    Angry/
        *.wav
    ...
"""

from pathlib import Path

import click

from ertk.dataset import resample_audio, write_annotations, write_filelist

emotion_map = {
    "A": "anger",
    "S": "sadness",
    "H": "happiness",
    "N": "neutral",
}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the URDU dataset at location INPUT_DIR and resample audio
    to 16 kHz 16-bit WAV audio.
    """

    paths = list(input_dir.glob("*/*.wav"))
    if resample:
        resample_dir = Path("resampled")
        resample_audio(paths, resample_dir)
        write_filelist(resample_dir.glob("*.wav"), "files_all")

    write_annotations(
        {p.stem: emotion_map[p.stem[p.stem.rfind("_") + 1]] for p in paths}, "label"
    )
    write_annotations({p.stem: p.stem[: p.stem.index("_")] for p in paths}, "speaker")
    write_annotations({p.stem: "ur" for p in paths}, "language")


if __name__ == "__main__":
    main()



"""PyTorch utilities."""

from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchaudio

__all__ = [
    "get_activation",
    "get_loss",
    "load_tensorboard",
    "frame_tensor",
    "add_signal_awgn",
    "add_gaussian",
    "spectrogram",
]


def get_loss(loss: str, *args, **kwargs) -> nn.Module:
    """Get torch Module for activation function.

    Parameters
    ----------
    loss: str
        String representing the loss function.
    *args, **kwargs: optional
        Arguments and keyword arguments to pass to loss function
        module `__init__()`.

    Returns
    -------
    torch.nn.Module
        The loss function as a `Module`.
    """
    cls = {
        "l1": nn.L1Loss,
        "mse": nn.MSELoss,
        "ce": nn.CrossEntropyLoss,
        "cross_entropy": nn.CrossEntropyLoss,
        "kl": nn.KLDivLoss,
        "bce": nn.BCELoss,
        "bce_logits": nn.BCEWithLogitsLoss,
        "ctc": nn.CTCLoss,
        "nll": nn.NLLLoss,
        "hinge": nn.HingeEmbeddingLoss,
        "cosine": nn.CosineEmbeddingLoss,
        "triplet": nn.TripletMarginLoss,
    }[loss]
    return cls(*args, **kwargs)


def get_activation(act: str, *args, **kwargs) -> nn.Module:
    """Get torch Module for activation function.

    Parameters
    ----------
    act: str
        String representing the activation function.
    *args, **kwargs: optional
        Arguments and keyword arguments to pass to activation function
        module `__init__()`.

    Returns
    -------
    torch.nn.Module
        The activation function as a `Module`.
    """
    cls = {
        "relu": nn.ReLU,
        "leaky_relu": nn.LeakyReLU,
        "glu": nn.GLU,
        "mish": nn.Mish,
        "selu": nn.SELU,
        "silu": nn.SiLU,
        "gelu": nn.GELU,
        "celu": nn.CELU,
        "prelu": nn.PReLU,
        "rrelu": nn.RReLU,
        "elu": nn.ELU,
        "sigmoid": nn.Sigmoid,
        "log_sigmoid": nn.LogSigmoid,
        "tanh": nn.Tanh,
        "threshold": nn.Threshold,
        "softmax": nn.Softmax,
        "log_softmax": nn.LogSoftmax,
        "softplus": nn.Softplus,
        "linear": nn.Identity,
    }[act]
    return cls(*args, **kwargs)


def load_tensorboard():
    """Replaces the TensorFlow tensorboard with the tensorboard stub.
    This avoids issues with having both PyTorch and TensorFlow installed
    simultaneously.
    """
    import tensorboard.compat.tensorflow_stub.io.gfile as _gfile
    import tensorflow as _tensorflow

    _tensorflow.io.gfile = _gfile


def frame_tensor(
    x: torch.Tensor,
    frame_size: int,
    frame_shift: int,
    pad: bool = False,
    axis: int = 0,
) -> torch.Tensor:
    """Frames an array over a given axis with optional padding and
    copying.

    Parameters
    ----------
    x: torch.Tensor
        Array to frame.
    frame_size: int
        Width of frame.
    frame_shift: int
        Amount each window is moved for subsequent frames.
    pad: bool
        If the frames don't neatly fit the axis length, this indicated
        whether to pad or cut the remaining length. Default is `False`
        which means any remaining length is cut. If `pad=True` then any
        remaining length is padded with zeros and an additional frame is
        produced. If frames neatly fit into the axis length then this
        option has no effect.
    axis: int
        Axis to frame over. Default is 0.

    Returns
    -------
    frames: torch.Tensor
        The resulting frames.
    """

    if axis < 0:
        axis += x.ndim

    L = x.size(axis)
    if L < frame_size and not pad:
        raise ValueError(
            "The length of the sequence is shorter than frame_size, but pad=False, "
            "so no frames will be generated."
        )

    num_frames = (L - frame_size) // frame_shift + 1
    remainder = (L - frame_size) % frame_shift
    if remainder != 0 and pad:
        if num_frames <= 0:
            left = L
        else:
            left = frame_size + remainder - frame_shift
        padding = (0, 0) * axis + (frame_size - left, 0) + (0, 0) * (x.ndim - axis - 1)
        x = F.pad(x, tuple(reversed(padding)))
    framed = x.unfold(axis, frame_size, frame_shift)
    if axis != framed.ndim - 2:
        return framed.transpose(axis + 1, framed.ndim - 1)
    return framed


def add_signal_awgn(x: torch.Tensor, snr: float = 20) -> torch.Tensor:
    """Add additive white Gaussian noise (AWGN) to a signal or batch of
    signals.

    Parameters
    ----------
    x: torch.Tensor
        Input tensor of shape (signal_len,) or (batch_size, signal_len).
    snr: float
        Target signal-to-noise ratio.

    Returns
    -------
    augmented: torch.Tensor
        Augmented tensor of the same shape as the `x`.
    """
    x = x.squeeze(-1)
    noise = torch.randn(x.size(), dtype=x.dtype, device=x.device)
    return torchaudio.functional.add_noise(x, noise, snr)


def add_gaussian(x: torch.Tensor, sigma: float = 1, multiplicative: bool = True):
    """Add Gaussian noise to a tensor or batch of tensors.

    Parameters
    ----------
    x: torch.Tensor
        Input tensor of any shape.
    sigma: float
        Standard deviation of Gaussian.
    multiplicative: bool
        Additive or multiplicative noise.

    Returns
    -------
    augmented: torch.Tensor
        Augmented tensor of the same shape as the `x`.
    """
    noise = sigma * torch.randn(x.size(), dtype=x.dtype, device=x.device)
    if multiplicative:
        return x * (1 + noise)
    else:
        return x + noise


def spectrogram(
    x: torch.Tensor,
    sr: float,
    kind: str = "mel",
    pre_emphasis: float = 0,
    window_size: float = 0.025,
    window_shift: float = 0.01,
    n_fft: int = 2048,
    n_mels: int = 128,
    htk_mel: bool = False,
    n_chroma: int = 12,
    clip_db: Optional[float] = None,
    fmin: float = 0,
    fmax: Optional[float] = 8000,
    power: int = 2,
    to_db: bool = True,
):
    window_samples = int(window_size * sr)
    stride_samples = int(window_shift * sr)
    if kind == "mel":
        module = torchaudio.transforms.MelSpectrogram(
            sample_rate=sr,
            n_fft=n_fft,
            win_length=window_samples,
            hop_length=stride_samples,
            f_min=fmin,
            f_max=fmax,
            n_mels=n_mels,
            power=power,
            mel_scale="htk" if htk_mel else "slaney",
        )
    else:
        module = torchaudio.transforms.Spectrogram(
            n_fft=n_fft,
            win_length=window_samples,
            hop_length=stride_samples,
            power=power,
        )
    with torch.no_grad():
        spec = module(x).transpose(-1, -2)
        if to_db:
            spec = torchaudio.transforms.AmplitudeToDB(
                stype="power" if power == 2 else "magnitude", top_db=clip_db
            )(spec)



"""Process the raw MESD dataset.

This assumes the file structure from the original dataset:
/.../
    Anger_C_A_abajo.wav
    Anger_C_A_adios.wav
    ...
"""

from pathlib import Path

import click

from ertk.dataset import resample_audio, write_annotations, write_filelist


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the MESD dataset at location INPUT_DIR and resample audio
    to 16 kHz 16-bit WAV audio.
    """

    paths = list(input_dir.glob("*.wav"))
    resample_dir = Path("resampled")
    if resample:
        resample_audio(paths, resample_dir)
    write_filelist(resample_dir.glob("*.wav"), "files_all")

    write_annotations({p.stem: p.stem.split("_")[0].lower() for p in paths}, "label")
    write_annotations({p.stem: p.stem.split("_", maxsplit=3)[3] for p in paths}, "word")
    write_annotations({p.stem: p.stem.split("_")[1] for p in paths}, "voice")
    write_annotations({p.stem: p.stem.split("_")[2] for p in paths}, "mesd_corpus")
    write_annotations({p.stem: "es" for p in paths}, "language")
    write_annotations({p.stem: "mx" for p in paths}, "country")


if __name__ == "__main__":
    main()



import json
import warnings
from pathlib import Path
from typing import Tuple

import click
import numpy as np
import pandas as pd
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler, StandardScaler

from ertk.classification import (
    dataset_cross_validation,
    dataset_train_val_test,
    get_balanced_sample_weights,
)
from ertk.cli._utils import (
    dataset_args,
    debug_args,
    eval_args,
    model_args,
    result_args,
    train_args,
)
from ertk.config import get_arg_mapping
from ertk.dataset import load_multiple
from ertk.sklearn.utils import GridSearchVal
from ertk.train import ValidationSplit, get_cv_splitter
from ertk.transform import SequenceTransformWrapper


@click.command()
@dataset_args
@eval_args
@model_args
@result_args
@train_args
@debug_args
def main(
    clf_type: str,
    corpus_info: Tuple[Path],
    data_config: Path,
    features: str,
    cv_part: str,
    label: str,
    kfold: int,
    inner_kfold: int,
    test_size: float,
    inner_part: str,
    results: Path,
    logdir: Path,
    reps: int,
    normalise: str,
    transform: str,
    balanced: bool,
    subset: Tuple[str],
    map_groups: Tuple[str],
    sel_groups: Tuple[str],
    remove_groups: Tuple[str],
    clip_seq: int,
    pad_seq: int,
    verbose: int,
    clf_args_file: Tuple[Path],
    param_grid_file: Tuple[Path],
    learning_rate: float,
    batch_size: int,
    epochs: int,
    use_inner_cv: bool,
    n_jobs: int,
    n_gpus: int,
    train: str,
    valid: str,
    test: str,
    train_config_path: Path,
    seq_transform: str,
    sample_rate: int,
):
    np.set_printoptions(precision=4, threshold=10)
    if verbose > 0:
        import logging

        logging.basicConfig(level=logging.DEBUG if verbose > 1 else logging.INFO)
        np.set_printoptions()

    if (train and kfold) or (not train and not kfold):
        raise ValueError("Exactly one of train and kfold must be given.")

    read_kwargs = {"sample_rate": sample_rate} if sample_rate else {}
    if len(subset) == 1 and ":" not in subset[0]:
        dataset = load_multiple(
            corpus_info, features, subsets=subset[0], label=label, **read_kwargs
        )
    else:
        subset_map = {}
        for m in subset:
            subset_map.update(get_arg_mapping(m))
        dataset = load_multiple(
            corpus_info, features, subsets=subset_map, label=label, **read_kwargs
        )

    grp_map = {}
    for m in map_groups:
        grp_map.update(get_arg_mapping(m))
    grp_sel = {}
    for m in sel_groups:
        grp_sel.update(get_arg_mapping(m))
    grp_del = {}
    for m in remove_groups:
        grp_del.update(get_arg_mapping(m))
    dataset.map_and_select(grp_map, grp_sel, grp_del)

    if clip_seq:
        dataset.clip_arrays(clip_seq)
    if pad_seq:
        dataset.pad_arrays(pad_seq)

    transformer = {"std": StandardScaler, "minmax": MinMaxScaler}[transform]()
    if normalise == "none":
        transformer = None
    elif normalise == "online" and len(dataset.x[0].shape) > 1:
        transformer = SequenceTransformWrapper(transformer, seq_transform)
    elif normalise != "online":
        dataset.normalise(normaliser=transformer, partition=normalise)
        transformer = None

    sample_weight = None
    if balanced:
        sample_weight = get_balanced_sample_weights(dataset.get_annotations(label))
    fit_params = {"sample_weight": sample_weight}

    model_args = {}
    for file in clf_args_file:
        model_args.update(get_arg_mapping(file))
    param_grid = {}
    for file in param_grid_file:
        param_grid.update(get_arg_mapping(file))

    if not train:
        cv = get_cv_splitter(bool(cv_part), kfold, test_size=test_size)
        if not use_inner_cv:
            warnings.warn(
                "--noinner_cv is deprecated and only exists for backwards "
                "compatibility."
            )
            inner_cv = cv
        else:
            inner_cv = get_cv_splitter(
                bool(inner_part), inner_kfold, shuffle=True, random_state=54321
            )
    else:
        if not valid:
            raise ValueError("valid must be specified.")
        train_indices = dataset.get_idx_for_split(train)
        valid_indices = dataset.get_idx_for_split(valid)
        if not test:
            test = valid
        test_indices = dataset.get_idx_for_split(test)
        inner_cv = ValidationSplit(
            np.arange(len(train_indices), len(train_indices) + len(valid_indices))
        )

    clf_lib, clf_type = clf_type.split("/")
    if clf_lib == "sk":
        from sklearn.model_selection import GridSearchCV

        from ertk.sklearn.models import get_sk_model

        clf = get_sk_model(clf_type, **model_args)
        if train:
            # A hack so that the ValidationSplit works properly.
            train_indices = np.concatenate([train_indices, valid_indices])
            grid_search = GridSearchVal
        else:
            grid_search = GridSearchCV
        clf = grid_search(
            Pipeline([("transform", transformer), ("clf", clf)]),
            {f"clf__{k}": v for k, v in param_grid.items()},
            scoring="balanced_accuracy",
            cv=inner_cv,
            verbose=max(verbose, 0),
            n_jobs=1 if use_inner_cv and not train else n_jobs,
        )
        if inner_part:
            fit_params["groups"] = dataset.get_group_indices(inner_part)
        if not use_inner_cv:
            # Get best params then pass best to main cross-validation routine
            # XXX: This exists because of mistake in our original implementation.
            clf.fit(
                dataset.x,
                dataset.labels,
                groups=dataset.get_group_indices(cv_part) if cv_part else None,
                clf__sample_weight=sample_weight,
            )
            params = clf.best_params_
            clf = clf.best_estimator_
    elif clf_lib == "tf":
        from keras.callbacks import TensorBoard
        from keras.optimizers import get as get_optimizer

        from ertk.tensorflow import get_tf_model
        from ertk.tensorflow.classification import tf_classification_metrics

        callbacks = []
        if logdir is not None:
            callbacks.append(
                TensorBoard(
                    log_dir=logdir,
                    profile_batch=0,
                    write_graph=False,
                    write_images=False,
                )
            )
        # No parallel CV to avoid running out of GPU memory
        n_jobs = 1
        fit_params.update(
            {
                "epochs": epochs,
                "callbacks": callbacks,
                "batch_size": batch_size,
                "data_fn": None,
                "n_gpus": n_gpus,
                "log_dir": logdir,
            }
        )
        model_args.update(
            {"n_features": dataset.n_features, "n_classes": dataset.n_classes}
        )

        def model_fn():
            model = get_tf_model(clf_type, **model_args)
            model.compile(
                get_optimizer("adam")(learning_rate=learning_rate),
                loss="sparse_categorical_crossentropy",
                metrics=tf_classification_metrics(),
            )
            return Pipeline([("transform", transformer), ("clf", model)])

        params = {
            "optimiser": "adam",
            "learning_rate": learning_rate,
            "batch_size": batch_size,
            **model_args,
            **fit_params,
        }
        clf = model_fn
    elif clf_lib == "pt":
        from omegaconf import OmegaConf

        from ertk.pytorch.models import PyTorchModelConfig, get_pt_model
        from ertk.pytorch.train import PyTorchLoggingConfig, PyTorchTrainConfig

        n_jobs = 1
        if train_config_path:
            train_config = PyTorchTrainConfig.from_file(train_config_path)
        else:
            train_config = PyTorchTrainConfig(
                batch_size=batch_size,
                n_gpus=n_gpus,
                epochs=epochs,
                logging=PyTorchLoggingConfig(log_dir=str(logdir)),
            )
        fit_params.update({"train_config": train_config})

        model_config: PyTorchModelConfig = OmegaConf.create(model_args)
        if model_config.n_features == -1:
            model_config.n_features = dataset.n_features
        elif model_config.n_features != dataset.n_features:
            logging.warning("`n_features` differs between dataset and model_config")
        if model_config.n_classes == -1:
            model_config.n_classes = dataset.n_classes
        elif model_config.n_classes != dataset.n_classes:
            raise ValueError("`n_classes` differs between dataset and model_config")

        def model_fn():
            model = get_pt_model(clf_type, config=model_config)
            return model if transformer is None else (transformer, model)

        params = {
            **PyTorchModelConfig.to_dictconfig(model_config),
            **PyTorchTrainConfig.to_dictconfig(train_config),
        }
        clf = model_fn
    else:
        raise ValueError(f"Invalid classifier type {clf_type}")

    if verbose > -1:
        print("== Dataset ==")
        print(dataset)
        print("== Training/evaluation settings ==")
        print(f"Using {dataset.n_classes}-class classification.")
        print(f"Using {n_jobs} parallel jobs.")
        print(f"Using classifier {clf}.")
        print(f"Using fit_params={fit_params}")
        if train:
            print(f"Training set: {train}")
            print(f"Validation set: {valid}")
            print(f"Test set: {test}")
        else:
            print(f"Using {kfold} k-fold CV.")
            if cv_part:
                print(f"Using {cv_part} as split partition.")
            if use_inner_cv:
                if inner_part:
                    print(f"Using {inner_part} as inner split partition.")
                print(f"Using {inner_cv} as inner CV splitter.")
        if balanced:
            print("Using balanced sample weights.")
        if normalise == "online":
            print("Using 'online' normalisation.")
        elif normalise != "none":
            print(f"Normalising globally using {normalise} partition.")
        if normalise != "none":
            print(f"Using transformer {transformer}")

    dfs = []
    for rep in range(1, reps + 1):
        if verbose > -1:
            print(f"Rep {rep}/{reps}")
        if train:
            res = dataset_train_val_test(
                clf,
                dataset=dataset,
                train_idx=train_indices,
                valid_idx=valid_indices,
                test_idx=test_indices,
                label=label,
                clf_lib=clf_lib,
                sample_weight=sample_weight,
                verbose=verbose,
                fit_params=fit_params,
            )
        else:
            res = dataset_cross_validation(
                clf,
                dataset,
                clf_lib=clf_lib,
                partition=cv_part,
                label=label,
                cv=cv,
                verbose=max(verbose, 0),
                n_jobs=n_jobs,
                fit_params=fit_params,
            )
        df = res.scores_df
        if "params" not in df:
            df["params"] = [json.dumps(params, default=str)] * len(df)
        dfs.append(df)
    df = pd.concat(dfs, keys=list(range(1, reps + 1)), names=["rep"])
    df["clf"] = f"{clf_lib}/{clf_type}"
    df["features"] = features
    df["corpus"] = dataset.corpus
    if results:
        results.parent.mkdir(parents=True, exist_ok=True)
        df.to_csv(results)
        print(f"Wrote CSV to {results}")
    else:
        print(df)
    if verbose > -1:
        print(df.mean(numeric_only=True).to_string())


if __name__ == "__main__":
    main()



"""PyTorch implementations of various models.

.. autosummary::
    :toctree:

    get_pt_model
    get_pt_model_fn
    PyTorchModelConfig
    ERTKPyTorchModel
    SimpleModel
    SimpleClassificationModel
    LightningWrapper


Models
------
.. autosummary::
    :toctree:

    aldeneh2017
    mlp
    zhang2019


Layers
------
.. autosummary::
    :toctree:

    layers
"""

from functools import partial
from typing import Callable

from torch.nn import Module

# Needed for registration
from . import aldeneh2017, mlp, wav2vec_ft, zhang2019
from ._base import (
    ERTKPyTorchModel,
    LightningWrapper,
    PyTorchModelConfig,
    SimpleClassificationModel,
    SimpleModel,
)

__all__ = [
    "ERTKPyTorchModel",
    "LightningWrapper",
    "PyTorchModelConfig",
    "SimpleClassificationModel",
    "SimpleModel",
    "get_pt_model",
    "get_pt_model_fn",
]


def get_pt_model_fn(name: str, **kwargs) -> Callable[..., Module]:
    """Get a PyTorch model creation function.

    Parameters
    ----------
    name: str
        The model name. This should be the name of the module that
        implements the model.
    **kwargs:
        Keyword arguments to provide to the model creation function.

    Returns
    -------
    model_fn: callable
        A method that takes arguments and returns a Module instance.
    """
    model_cls = ERTKPyTorchModel.get_model_class(name)
    return partial(model_cls, **kwargs)


def get_pt_model(name: str, **kwargs) -> Module:
    """Get a PyTorch model by name.

    Parameters
    ----------
    name: str
        The model name. This should be the name of the module that
        implements the model.
    **kwargs:
        Keyword arguments to provide to the model creation function.

    Returns
    -------
    model: Module
        The PyTorch model.
    """
    return get_pt_model_fn(name, **kwargs)()



"""Kmeans vector quantiser."""

from dataclasses import dataclass
from typing import List

import numpy as np
from omegaconf import MISSING

from ertk.config import ERTKConfig

from ._base import FeatureExtractor

__all__ = ["KMeansExtractorConfig", "KMeansExtractor"]


@dataclass
class KMeansExtractorConfig(ERTKConfig):
    """KMeans feature extractor configuration."""

    pickle: str = MISSING
    """Path to pickle file."""


class KMeansExtractor(FeatureExtractor, fname="kmeans", config=KMeansExtractorConfig):
    """KMeans vector quantiser."""

    config: KMeansExtractorConfig

    def __init__(self, config: KMeansExtractorConfig) -> None:
        super().__init__(config)

        import joblib

        self.kmeans = joblib.load(config.pickle)

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        return np.expand_dims(self.kmeans.predict(x), axis=1)

    @property
    def is_sequence(self) -> bool:
        return True

    @property
    def feature_names(self) -> List[str]:
        return ["cluster_id"]



"""Functions for reading and writing annotations."""

from typing import Dict, Mapping, Optional, Type, Union

import pandas as pd
from typing_extensions import Literal

from ertk.utils import PathOrStr

__all__ = ["read_annotations", "write_annotations"]


def read_annotations(
    filename: PathOrStr, dtype: Optional[Union[Type, Literal["category"]]] = None
) -> pd.Series:
    """Returns a pd.Series containing values for this annotation for
    each instance, indexed by name.
    """
    _dtype = dtype
    if _dtype == "category":
        _dtype = str
    dtypes: Dict[int, Type] = {0: str}
    if _dtype is not None:
        dtypes[1] = _dtype
    df = pd.read_csv(filename, index_col=0, header=0, dtype=dtypes, low_memory=False)
    if dtype == "category":
        return df[df.columns[0]].astype("category")
    return df[df.columns[0]]


def write_annotations(
    annotations: Union[Mapping[str, object], pd.DataFrame, pd.Series],
    name: Optional[str] = None,
    path: Optional[PathOrStr] = None,
) -> None:
    """Write sorted annotations CSV.

    Parameters
    ----------
    annotations: mapping
        A mapping of the form {name: annotation}.
    name: str, optional
        Name of the annotation if passing a dict.
    path: pathlike or str, optional
        Path to write CSV. If None, filename is name.csv
    """
    if isinstance(annotations, pd.DataFrame):
        df = annotations
        if len(df.columns) == 2:
            try:
                df.set_index("name", inplace=True)
            except KeyError:
                raise ValueError("Passed DataFrame should have 'name' as index.")
        elif len(df.columns) > 2:
            raise ValueError("Passed DataFrame should have 1 or 2 columns.")
        series = df[df.columns[0]]
    elif isinstance(annotations, pd.Series):
        series = annotations
    else:
        if not name:
            raise ValueError("`name` must be given when passing a dict.")
        df = pd.DataFrame.from_dict(annotations, orient="index", columns=[name])
        series = df[name]
    series.index.name = "name"
    name = name or series.name
    series.name = name
    path = path or f"{name}.csv"
    series.sort_index().to_csv(path, header=True, index=True)



"""Process the raw SUBESCO dataset.

This assumes the file structure from the original compressed file:
/.../
    Colère/
        Faible/
            *.aiff [for 192k data]
            *.wav  [for 48k data]
        Fort/
    ...
"""

import re
from pathlib import Path
from typing import Dict

import click

from ertk.dataset import resample_audio, write_annotations, write_filelist

REGEX = re.compile(r"^([MF])_([0-9]+)_([A-Z]+)_S_([0-9]+)_([A-Z]+)_([0-9])$")

emotion_map = {
    "ANGRY": "anger",
    "DISGUST": "disgust",
    "HAPPY": "happiness",
    "NEUTRAL": "neutral",
    "FEAR": "fear",
    "SURPRISE": "surprise",
    "SAD": "sadness",
}

sentence_map = {
    1: " েমৗমািছর চাকেদেখ কুকুরিট েঘউেঘউ করেছ",
    2: "তু িমসবউলটপালট কের িদেয়ছ",
    3: " েসেকানিকছু নাবেলই চেল েগেছ",
    4: " েতামােক এখিন আমার সােথ েযেত হেব",
    5: " েতামার কাজটা করািঠকহয়িন",
    6: "ইঁদুর ছানাটা হািরেয় েগল",
    7: "এখন পৰ্শ্ন করাযােব না",
    8: "দরজার বাইের কুকুরিট দাঁিড়েয় আেছ",
    9: "একিদন পেরই তারিবেয়",
    10: "ডাকােতরা ঢালতেলায়ার িনেয় এল",
}


@click.command()
@click.argument(
    "input_dir", type=click.Path(exists=True, file_okay=False, path_type=Path)
)
@click.option("--resample/--noresample", default=True)
def main(input_dir: Path, resample: bool):
    """Process the SUBESCO dataset at location INPUT_DIR and resample
    audio to 16 kHz 16-bit WAV audio.
    """

    paths = list(input_dir.glob("*.wav"))
    if resample:
        resample_dir = Path("resampled")
        resample_audio(paths, resample_dir)
        write_filelist(resample_dir.glob("*.wav"), "files_all")

    keys = ["label", "sentence", "speaker", "speaker_name", "gender", "transcript"]
    annot: Dict[str, Dict[str, str]] = {x: {} for x in keys}
    for p in paths:
        match = REGEX.match(p.stem)
        if not match:
            print(p)
        else:
            annot["gender"][p.stem] = match[1]
            annot["speaker"][p.stem] = match[2]
            annot["speaker_name"][p.stem] = match[3]
            annot["sentence"][p.stem] = match[4]
            annot["transcript"][p.stem] = sentence_map[int(match[4])]
            annot["label"][p.stem] = emotion_map[match[5]]

    for k in keys:
        write_annotations(annot[k], k)
    write_annotations({p.stem: "bn" for p in paths}, "language")


if __name__ == "__main__":
    main()



"""PyTorch dataset utilities."""

from typing import Any, Callable, Collection, Dict, List, Optional, Tuple, Union

import numpy as np
import pytorch_lightning as pl
import torch
from torch.utils.data import DataLoader
from torch.utils.data import Dataset as TorchDataset
from torch.utils.data import TensorDataset

from ertk.dataset import Dataset, load_datasets_config
from ertk.train import ExperimentConfig

__all__ = [
    "DatasetFuncWrapper",
    "DataModuleAdapter",
    "MTLDataModule",
    "MTLDataset",
    "create_mtl_dataloader",
]


class DatasetFuncWrapper(TorchDataset[Tuple[torch.Tensor, ...]]):
    funcs: List[Callable[[torch.Tensor], torch.Tensor]]

    def __init__(self, dataset: TorchDataset) -> None:
        super().__init__()
        self.dataset = dataset
        self.funcs = []

    def add_func(self, func: Callable[[torch.Tensor], torch.Tensor]):
        self.funcs.append(func)

    def __getitem__(self, index: Any) -> Tuple[torch.Tensor, ...]:
        x, *rest = self.dataset[index]
        for func in self.funcs:
            x = func(x)
        return (x,) + tuple(rest)

    def __len__(self) -> int:
        if hasattr(self.dataset, "__len__"):
            return len(self.dataset)  # type: ignore
        else:
            return NotImplemented


class DataModuleAdapter(pl.LightningDataModule):
    train_dataset: TorchDataset
    val_dataset: TorchDataset
    test_dataset: TorchDataset

    def __init__(
        self,
        *,
        load_config: Optional[ExperimentConfig] = None,
        dataset: Optional[Dataset] = None,
        train_select: Union[str, Dict[str, Collection[str]], np.ndarray, None] = None,
        val_select: Union[str, Dict[str, Collection[str]], np.ndarray, None] = None,
        test_select: Union[str, Dict[str, Collection[str]], np.ndarray, None] = None,
        batch_size: int = 32,
        dl_num_workers: int = 0,
    ):
        super().__init__()

        self.load_config = load_config
        self.dataset = dataset
        self.batch_size = batch_size
        self.train_select = train_select
        self.val_select = val_select
        self.test_select = test_select if test_select is not None else val_select
        self.dl_num_workers = dl_num_workers

    def setup(self, stage: Optional[str] = None) -> None:
        if self.load_config:
            data = load_datasets_config(self.load_config)
            self.train_idx = data.get_idx_for_split(self.load_config.eval.train)
            self.val_idx = data.get_idx_for_split(self.load_config.eval.valid)
            self.test_idx = data.get_idx_for_split(self.load_config.eval.test)

        if isinstance(self.train_select, np.ndarray):
            self.train_idx = self.train_select
        else:
            self.train_idx = self.dataset.get_idx_for_split(self.train_select)
        if isinstance(self.val_select, np.ndarray):
            self.val_idx = self.val_select
        else:
            self.val_idx = self.dataset.get_idx_for_split(self.val_select)
        if isinstance(self.test_select, np.ndarray):
            self.test_idx = self.test_select
        else:
            self.test_idx = self.dataset.get_idx_for_split(self.test_select)

        self._create_datasets()

    def _create_datasets(self) -> None:
        train_x = torch.from_numpy(self.dataset.x[self.train_idx])
        train_y = torch.from_numpy(self.dataset.y[self.train_idx])
        self.train_dataset = TensorDataset(train_x, train_y)
        val_x = torch.from_numpy(self.dataset.x[self.val_idx])
        val_y = torch.from_numpy(self.dataset.y[self.val_idx])
        self.val_dataset = TensorDataset(val_x, val_y)
        test_x = torch.from_numpy(self.dataset.x[self.test_idx])
        test_y = torch.from_numpy(self.dataset.y[self.test_idx])
        self.test_dataset = TensorDataset(test_x, test_y)

    def train_dataloader(self) -> DataLoader:
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.dl_num_workers,
        )

    def val_dataloader(self) -> DataLoader:
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.dl_num_workers,
        )

    def test_dataloader(self) -> DataLoader:
        return DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.dl_num_workers,
        )

    def predict_dataloader(self) -> DataLoader:
        return DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.dl_num_workers,
        )


class MTLDataModule(DataModuleAdapter):
    def __init__(
        self,
        *,
        dataset: Dataset,
        tasks: List[str],
        train_select: Union[str, Dict[str, Collection[str]], np.ndarray, None] = None,
        val_select: Union[str, Dict[str, Collection[str]], np.ndarray, None] = None,
        test_select: Union[str, Dict[str, Collection[str]], np.ndarray, None] = None,
        batch_size: int = 32,
        dl_num_workers: int = 0,
    ):
        super().__init__(
            dataset=dataset,
            train_select=train_select,
            val_select=val_select,
            test_select=test_select,
            batch_size=batch_size,
            dl_num_workers=dl_num_workers,
        )
        self.tasks = tasks

    def _create_datasets(self) -> None:
        for name in ["train", "val", "test"]:
            idx = getattr(self, f"{name}_idx")
            x = torch.as_tensor(self.dataset.x[idx])
            ys = {
                k: torch.as_tensor(
                    self.dataset.get_group_indices(k)[idx], dtype=torch.long
                )
                for k in self.tasks
            }
            setattr(self, f"{name}_dataset", MTLDataset(x, ys))


class MTLDataset(TorchDataset[Tuple[torch.Tensor, Dict[str, torch.Tensor]]]):
    def __init__(self, x: torch.Tensor, y: Dict[str, torch.Tensor]):
        self.x = x
        self.y = y

    def __getitem__(self, index):
        return self.x[index], {k: v[index] for k, v in self.y.items()}

    def __len__(self):
        return self.x.size(0)


def create_mtl_dataloader(
    data: Dataset, tasks: List[str], batch_size: int = 32, shuffle: bool = False
) -> DataLoader[Tuple[torch.Tensor, Dict[str, torch.Tensor]]]:
    x = torch.as_tensor(data.x)
    task_data = {k: torch.as_tensor(data.get_group_indices(k)) for k in tasks}
    dataset = MTLDataset(x, task_data)
    return DataLoader(dataset, batch_size, shuffle=shuffle)



"""Implementation of auDeep model from [1]_.

References
----------
.. [1] M. Freitag, S. Amiriparian, S. Pugachevskiy, N. Cummins, and B.
       Schuller, “auDeep: Unsupervised learning of representations from
       audio with deep recurrent neural networks,” The Journal of
       Machine Learning Research, vol. 18, no. 1, pp. 6340-6344, 2017.
"""

import time
from pathlib import Path
from typing import Any, Dict, Tuple, Union

import click
import numpy as np
import torch
import torch.nn.functional as F
from torch import nn
from torch.optim import Adam
from torch.utils.data import DataLoader, TensorDataset
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from ertk.dataset import read_features, write_features
from ertk.utils import PathOrStr

__all__ = ["TimeRecurrentAutoencoder"]


class TimeRecurrentAutoencoder(nn.Module):
    """Time recurrent autoencoder model.

    Parameters
    ----------
    n_features : int
        Number of features in the input data.
    n_units : int, optional
        Number of units in each layer of the encoder and decoder.
    n_layers : int, optional
        Number of layers in the encoder and decoder.
    dropout : float, optional
        Dropout probability.
    bidirectional_encoder : bool, optional
        Whether to use a bidirectional encoder.
    bidirectional_decoder : bool, optional
        Whether to use a bidirectional decoder.
    """

    def __init__(
        self,
        n_features: int,
        n_units: int = 256,
        n_layers: int = 2,
        dropout: float = 0.2,
        bidirectional_encoder: bool = False,
        bidirectional_decoder: bool = False,
    ):
        super().__init__()

        encoder_state_size = n_units * n_layers
        decoder_state_size = n_units * n_layers
        if bidirectional_encoder:
            encoder_state_size *= 2
        if bidirectional_decoder:
            decoder_state_size *= 2
        decoder_output_size = n_units * n_layers  # concat(fwd, back)
        self.embedding_size = decoder_state_size

        self.n_features = n_features
        self.n_units = n_units
        self.n_layers = n_layers
        self.dropout = dropout
        self.bidirectional_encoder = bidirectional_encoder
        self.bidirectional_decoder = bidirectional_decoder

        self.encoder = nn.GRU(
            n_features,
            n_units,
            num_layers=n_layers,
            dropout=dropout,
            bidirectional=bidirectional_encoder,
            batch_first=True,
        )
        self.representation = nn.Sequential(
            nn.Linear(encoder_state_size, decoder_state_size), nn.Tanh()
        )
        self.decoder = nn.GRU(
            n_features,
            n_units,
            num_layers=n_layers,
            dropout=dropout,
            bidirectional=bidirectional_decoder,
            batch_first=True,
        )
        self.reconstruction = nn.Sequential(
            nn.Linear(decoder_output_size, n_features), nn.Tanh()
        )

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        # x (B, T, F)
        targets = x.flip(1)

        num_directions = 2 if self.bidirectional_encoder else 1
        encoder_h = torch.zeros(
            self.n_layers * num_directions, x.size(0), self.n_units, device=x.device
        )
        _, encoder_h = self.encoder(
            F.dropout(x, self.dropout, self.training), encoder_h
        )
        encoder_h = torch.cat(encoder_h.unbind(), -1)

        representation = self.representation(encoder_h)

        num_directions = 2 if self.bidirectional_decoder else 1
        decoder_h = torch.stack(
            representation.chunk(self.n_layers * num_directions, -1)
        )

        decoder_input = F.pad(targets[:, :-1, :], [0, 0, 1, 0])
        decoder_seq, _ = self.decoder(
            F.dropout(decoder_input, self.dropout, self.training), decoder_h
        )
        decoder_seq = F.dropout(decoder_seq, self.dropout, self.training)

        reconstruction = self.reconstruction(decoder_seq)
        reconstruction = reconstruction.flip(1)
        return reconstruction, representation


def _get_latest_save(directory: Union[str, Path]) -> Path:
    """Gets the path to the latest model checkpoint, assuming it was saved with
    the format 'model-XXXX.pt'
    """
    directory = Path(directory)
    saves = list(directory.glob("model-*.pt"))
    if len(saves) == 0:
        raise FileNotFoundError(f"No save files found in directory {directory}")
    return max(saves, key=lambda p: int(p.stem.split("-")[1]))


def _save_model(
    directory: Union[Path, str],
    epoch: int,
    model: TimeRecurrentAutoencoder,
    model_args: Dict[str, Any],
    optimiser: Adam,
    optimiser_args: Dict[str, Any],
):
    """Saves the given model and associated training parameters."""
    save_path = Path(directory) / f"model-{epoch:04d}.pt"
    torch.save(
        {
            "epoch": epoch,
            "model_state_dict": model.state_dict(),
            "model_args": model_args,
            "optimiser_state_dict": optimiser.state_dict(),
            "optimiser_args": optimiser_args,
        },
        save_path,
    )
    print(f"Saved model to {save_path}.")


def _scale(x, low: float = -1, high: float = 1):
    xmin = np.min(x, axis=(1, 2), keepdims=True)
    xmax = np.max(x, axis=(1, 2), keepdims=True)
    return (high - low) * (x - xmin) / (xmax - xmin) + low


def _get_data(path: PathOrStr, shuffle: bool = False):
    """Get data and metadata from netCDF dataset. Optionally shuffle x."""
    dataset = read_features(path)
    x = dataset.features
    names = np.array(dataset.names)

    if shuffle:
        perm = np.random.default_rng().permutation(len(x))
        x = x[perm]
        names = names[perm]

    x = _scale(x, -1, 1)
    return x, names, dataset.corpus


@click.command()
@click.argument("input", type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.option("--logs", type=Path, required=True)
@click.option("--layers", type=int, default=2)
@click.option("--units", type=int, default=256)
@click.option("--dropout", type=float, default=0.2)
@click.option("--bidirectional_encoder", is_flag=True)
@click.option("--bidirectional_decoder", is_flag=True)
@click.option("--batch_size", type=int, default=64)
@click.option("--epochs", type=int, default=50)
@click.option("--learning_rate", type=float, default=0.001)
@click.option("--valid_fraction", type=float, default=0.1)
@click.option("--continue", "cont", is_flag=True)
@click.option("--save_every", type=int, default=20)
@click.option("--device", type=str, default="cuda")
def train(
    input: Path,
    logs: Path,
    layers: int,
    units: int,
    dropout: float,
    bidirectional_encoder: bool,
    bidirectional_decoder: bool,
    batch_size: int,
    epochs: int,
    learning_rate: float,
    valid_fraction: float,
    cont: bool,
    save_every: int,
    device: str,
):
    print("Reading input from {}".format(input))
    x = _get_data(input, shuffle=True)[0]
    x = torch.tensor(x)

    n_valid = int(valid_fraction * x.size(0))
    train_data = TensorDataset(x[n_valid:])
    valid_data = TensorDataset(x[:n_valid])
    train_dl = DataLoader(
        train_data, batch_size=batch_size, shuffle=True, pin_memory=True
    )
    valid_dl = DataLoader(valid_data, batch_size=batch_size, pin_memory=True)

    if cont:
        save_path = _get_latest_save(logs)
        load_dict = torch.load(save_path)
        model_args = load_dict["model_args"]
        optimiser_args = load_dict["optimiser_args"]
        initial_epoch = 1 + load_dict["epoch"]
    else:
        model_args = dict(
            n_features=x.size(2),
            n_units=units,
            n_layers=layers,
            dropout=dropout,
            bidirectional_encoder=bidirectional_encoder,
            bidirectional_decoder=bidirectional_decoder,
        )
        optimiser_args = dict(lr=learning_rate, eps=1e-5)
        initial_epoch = 1
    final_epoch = initial_epoch + epochs - 1

    model = TimeRecurrentAutoencoder(**model_args)
    model = model.to(device=device)
    optimiser = Adam(model.parameters(), **optimiser_args)

    if cont:
        print(f"Restoring model from {save_path}")
        model.load_state_dict(load_dict["model_state_dict"])
        optimiser.load_state_dict(load_dict["optimiser_state_dict"])

    logs.mkdir(parents=True, exist_ok=True)
    train_writer = SummaryWriter(logs / "train")
    valid_writer = SummaryWriter(logs / "valid")
    with torch.no_grad():
        batch = next(iter(train_dl))[0].to(device=device)
        train_writer.add_graph(model, input_to_model=batch)

    for epoch in range(initial_epoch, final_epoch + 1):
        start_time = time.perf_counter()
        model.train()
        train_losses = []
        for (x,) in tqdm(
            train_dl, desc=f"Epoch {epoch:03d}", unit="batch", leave=False
        ):
            x = x.to(device=device)
            optimiser.zero_grad()
            reconstruction, representation = model(x)

            loss = F.mse_loss(x, reconstruction, reduction="mean")
            train_losses.append(loss.item())

            loss.backward()
            # Clip gradients to [-2, 2]
            nn.utils.clip_grad_value_(model.parameters(), 2)
            optimiser.step()
        train_time = time.perf_counter() - start_time
        train_loss = np.mean(train_losses)

        model.eval()
        with torch.no_grad():
            valid_losses = []
            it = iter(valid_dl)

            # Get loss and write summaries for first batch
            x = next(it)[0].to(device=device)
            reconstruction, representation = model(x)
            loss = F.mse_loss(x, reconstruction, reduction="mean")
            valid_losses.append(loss.item())

            if epoch % 10 == 0:
                images = torch.cat([x, reconstruction], 2).unsqueeze(-1)
                images = (images + 1) / 2
                images = images[:20]
                valid_writer.add_images(
                    "reconstruction", images, global_step=epoch, dataformats="NHWC"
                )
                valid_writer.add_histogram(
                    "representation", representation, global_step=epoch
                )

            # Just get the loss for the rest, no summaries
            for (x,) in it:
                x = x.to(device=device)
                reconstruction, representation = model(x)
                loss = F.mse_loss(x, reconstruction, reduction="mean")
                valid_losses.append(loss.item())
            valid_loss = np.mean(valid_losses)

            train_writer.add_scalar("loss", train_loss, global_step=epoch)
            valid_writer.add_scalar("loss", valid_loss, global_step=epoch)

            end_time = time.perf_counter() - start_time
            mean_time = train_time / len(train_dl)
            print(
                f"Epoch {epoch:03d}: {end_time:.2f}s ({mean_time:.2f}s per "
                f"batch), train loss: {train_loss:.4f}, valid loss: "
                f"{valid_loss:.4f}"
            )

        if epoch % save_every == 0:
            _save_model(
                logs,
                epoch=epoch,
                model=model,
                model_args=model_args,
                optimiser=optimiser,
                optimiser_args=optimiser_args,
            )

    _save_model(
        logs,
        epoch=final_epoch,
        model=model,
        model_args=model_args,
        optimiser=optimiser,
        optimiser_args=optimiser_args,
    )


@click.command()
@click.argument("input", type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.argument("output", type=Path)
@click.option(
    "--model",
    "model_path",
    required=True,
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
@click.option("--batch_size", type=int, default=128)
@click.option("--device", type=str, default="cuda")
def generate(
    input: Path, output: Path, model_path: Path, batch_size: int, device: str = "cuda"
):
    if model_path.is_dir():
        save_path = _get_latest_save(model_path)
    else:
        save_path = model_path
    print(f"Restoring model from {save_path}")
    load_dict = torch.load(save_path)
    model = TimeRecurrentAutoencoder(**load_dict["model_args"])
    model.cuda()
    model.load_state_dict(load_dict["model_state_dict"])

    x, names, corpus = _get_data(input)

    representations = np.empty((len(x), model.embedding_size))
    with torch.no_grad():
        model.eval()
        for i in range(0, len(x), batch_size):
            batch = torch.tensor(x[i : i + batch_size], device=device)
            _, representation = model(batch)
            representations[i : i + batch_size, :] = representation.cpu()

    feature_names = [f"audeep{i + 1}" for i in range(representations.shape[-1])]
    write_features(
        output,
        corpus=corpus,
        names=names,
        features=representations,
        feature_names=feature_names,
    )
    print(f"Wrote netCDF4 file to {output}")


@click.group(no_args_is_help=True)
def main():
    import os

    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"
    # Workaround for having both PyTorch and TensorFlow installed.
    import tensorboard.compat.tensorflow_stub.io.gfile as _gfile
    import tensorflow as _tensorflow

    _tensorflow.io.gfile = _gfile


main.add_command(generate)
main.add_command(train)


if __name__ == "__main__":
    main()



from pathlib import Path

import click
import soundfile
from tqdm import tqdm

from ertk.dataset import get_audio_paths


@click.command()
@click.argument("input", type=click.Path(exists=True, path_type=Path))
@click.argument("output", type=Path)
@click.option(
    "--minlength",
    type=float,
    default=5,
    help="Minimum length (in seconds) of resulting speech clips.",
)
@click.option(
    "--maxlength",
    type=float,
    default=5,
    help="Maximum length (in seconds) of resulting speech clips.",
)
def main(input: Path, output: Path, minlength: float, maxlength: float):
    """Selects speech clips in a given length range."""
    clips = []
    if input.is_dir():
        paths = sorted(input.glob("**/*.wav"))
    else:
        paths = get_audio_paths(input)
    print(f"Found {len(paths)} files total.")
    for path in tqdm(paths, desc="Processing clips", disable=None):
        length = soundfile.info(path).duration
        if minlength < length < maxlength:
            clips.append(str(path))
    print(f"Found {len(clips)} valid clips.")

    output.parent.mkdir(parents=True, exist_ok=True)
    with open(output, "w") as fid:
        fid.write("\n".join(clips) + "\n")


if __name__ == "__main__":
    main()



"""Utility functions for TensorFlow models."""

from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import numpy as np
import tensorflow as tf
from keras.layers import Layer, Wrapper
from keras.losses import Loss
from keras.metrics import Metric
from keras.models import Model
from keras.optimizers import Optimizer
from keras.optimizers import get as get_optimiser
from sklearn.pipeline import Pipeline

__all__ = [
    "compile_wrap",
    "test_fit",
    "init_gpu_memory_growth",
    "print_linear_model_structure",
    "TFModelFunction",
]

TFModelFunction = Callable[..., Union[Model, Pipeline]]


def compile_wrap(
    model_fn: Optional[TFModelFunction] = None,
    opt_cls: Type[Optimizer] = get_optimiser("adam"),
    opt_kwargs: Dict[str, Any] = dict(learning_rate=0.0001),
    metrics: List[Union[str, Metric]] = ["sparse_categorical_accuracy"],
    loss: Union[str, Loss] = "sparse_categorical_crossentropy",
    **compile_kwargs,
):
    """Wrapper that takes a model creation function and gives a new
    function which returns a compiled model with the given compile
    parameters.

    Parameters
    ----------
    model_fn: callable, optional
        A method that returns an uncompiled model.
    opt_cls: type
        The Optimizer class to use.
    opt_kwargs: dict
        Keyword arguments to pass to opt_cls.
    metrics: list
        List of metrics to use.
    loss: Loss
        The loss function to use.
    **compile_kwargs: dict
        Other keyword arguments to pass to the model's compile() method.
    """

    def _wrapper(func: Callable[..., Model]):
        @wraps(func)
        def new_model_fn(*args, **kwargs) -> Model:
            model = func(*args, **kwargs)
            model.compile(
                optimizer=opt_cls(**opt_kwargs),
                metrics=metrics,
                loss=loss,
                **compile_kwargs,
            )
            return model

        return new_model_fn

    if model_fn is not None:
        return _wrapper(model_fn)

    return _wrapper


def test_fit(
    model_fn: TFModelFunction,
    input_size: Tuple[int, ...],
    *args,
    batch_size: int = 64,
    num_instances: int = 7000,
    **kwargs,
):
    """Tests the given model architecture/structure by training it on
    dummy data.

    Parameters
    ----------
    model_fn: callable
        Function that returns a Keras model. Called as model_fn(*args,
        **kwargs).
    input_size: tuple of int
        Input shape to the model. This is used to generate dummy data of
        the correct shape.
    *args
        Positional arguments to pass to model_fn().
    batch_size: int
        The batch size to use.
    num_instances: int
        The number of instances to generate.
    **kwargs
        Keyword arguments to pass to model_fn().
    """
    for gpu in tf.config.get_visible_devices("GPU"):
        tf.config.experimental.set_memory_growth(gpu, True)

    compiled_fn = compile_wrap(model_fn)
    model = compiled_fn(*args, n_classes=7, **kwargs)
    model.summary()

    valid = num_instances // 10
    rng = np.random.default_rng()
    x = rng.normal(size=(num_instances,) + input_size)
    y = rng.integers(7, size=num_instances)
    train_data = tf.data.Dataset.from_tensor_slices((x[valid:], y[valid:]))
    train_data = train_data.batch(batch_size)
    valid_data = tf.data.Dataset.from_tensor_slices((x[:valid], y[:valid]))
    valid_data = valid_data.batch(batch_size)
    model.fit(train_data, validation_data=valid_data, epochs=2, verbose=1)


def print_linear_model_structure(model: Model) -> None:
    """Prints the structure of a "sequential" model by listing the layer
    types and shapes in order.

    Parameters
    ----------
    model: Model
        The model to describe.
    """

    def print_inner(model: Layer, depth: int = 0):
        indent = "\t" * depth
        if not isinstance(model, Model):
            print(indent, model.name, model.output_shape)
            return

        for layer in model.layers:
            name = layer.name
            if name.startswith("tf_op_layer_"):
                name = name[12:]

            print(indent, name, layer.output_shape)
            if isinstance(layer, Model):
                print_inner(layer, depth + 1)
            elif isinstance(layer, Wrapper):
                print_inner(layer.layer, depth + 1)

    print_inner(model)


def init_gpu_memory_growth() -> None:
    """Sets TensorFlow to allocate memory on GPU as needed instead of
    all at once.
    """
    try:
        for gpu in tf.config.list_physical_devices("GPU"):
            tf.config.experimental.set_memory_growth(gpu, True)
    except RuntimeError as e:
        print(e)



import pytest
from importlib_metadata import entry_points

from ertk.preprocessing import InstanceProcessor


class TestPlugins:
    def test_plugin_loading(self):
        # Need entry points so that imports don't override each other
        eps = entry_points().select(group="ertk.processors")
        for ep in eps:
            cls = ep.load()
            assert issubclass(cls, InstanceProcessor)

    def test_invalid_plugin(self):
        with pytest.raises(KeyError):
            InstanceProcessor.get_processor_class("invalid")



"""Feature extractor using Keras applications."""

from dataclasses import dataclass
from types import ModuleType
from typing import Iterable, List, Optional, Union

import numpy as np
from omegaconf import MISSING
from scipy.ndimage import zoom

from ertk.config import ERTKConfig
from ertk.preprocessing.spectrogram import spectrogram
from ertk.utils import is_mono_audio

from ._base import AudioClipProcessor, FeatureExtractor

__all__ = ["KerasAppsExtractorConfig", "KerasAppsExtractor"]


@dataclass
class _ModelArgs:
    module: ModuleType
    funcname: str
    size: int


MODEL_MAP = {}


def _init_models():
    from keras.applications import (
        densenet,
        efficientnet,
        inception_resnet_v2,
        inception_v3,
        mobilenet,
        mobilenet_v2,
        mobilenet_v3,
        nasnet,
        resnet,
        vgg16,
        vgg19,
        xception,
    )

    global MODEL_MAP
    MODEL_MAP = {
        "densenet121": _ModelArgs(densenet, "DenseNet121", 224),
        "densenet169": _ModelArgs(densenet, "DenseNet169", 224),
        "densenet201": _ModelArgs(densenet, "DenseNet201", 224),
        "efficientnet_b0": _ModelArgs(efficientnet, "EfficientNetB0", 224),
        "efficientnet_b1": _ModelArgs(efficientnet, "EfficientNetB1", 240),
        "efficientnet_b2": _ModelArgs(efficientnet, "EfficientNetB2", 260),
        "efficientnet_b3": _ModelArgs(efficientnet, "EfficientNetB3", 300),
        "efficientnet_b4": _ModelArgs(efficientnet, "EfficientNetB4", 380),
        "efficientnet_b5": _ModelArgs(efficientnet, "EfficientNetB5", 456),
        "efficientnet_b6": _ModelArgs(efficientnet, "EfficientNetB6", 528),
        "efficientnet_b7": _ModelArgs(efficientnet, "EfficientNetB7", 600),
        "inception_resnet_v2": _ModelArgs(
            inception_resnet_v2, "InceptionResNetV2", 299
        ),
        "inception_v3": _ModelArgs(inception_v3, "InceptionV3", 299),
        "mobilenet": _ModelArgs(mobilenet, "MobileNet", 224),
        "mobilenet_v2": _ModelArgs(mobilenet_v2, "MobileNetV2", 224),
        "mobilenet_v3_large": _ModelArgs(mobilenet_v3, "MobileNetV3Large", 224),
        "mobilenet_v3_small": _ModelArgs(mobilenet_v3, "MobileNetV3Small", 224),
        "nasnet_large": _ModelArgs(nasnet, "NASNetLarge", 331),
        "nasnet_mobile": _ModelArgs(nasnet, "NASNetMobile", 224),
        "resnet50": _ModelArgs(resnet, "ResNet50", 224),
        "resnet50_v2": _ModelArgs(resnet, "ResNet50V2", 224),
        "resnet101": _ModelArgs(resnet, "ResNet101", 224),
        "resnet101_v2": _ModelArgs(resnet, "ResNet101V2", 224),
        "resnet152": _ModelArgs(resnet, "ResNet152", 224),
        "resnet152_v2": _ModelArgs(resnet, "ResNet152V2", 224),
        "vgg16": _ModelArgs(vgg16, "VGG16", 224),
        "vgg19": _ModelArgs(vgg19, "VGG19", 224),
        "xception": _ModelArgs(xception, "Xception", 299),
    }


@dataclass
class KerasAppsExtractorConfig(ERTKConfig):
    """Configuration for KerasAppsExtractor."""

    model: str = MISSING
    """Name of the model to use. See https://keras.io/api/applications/
    for the list of available models.
    """
    size: Optional[int] = None
    """Size of the input image. If not specified, the default size of the
    model will be used.
    """
    n_mels: int = 128
    """Number of mel bands for spectrogram extraction."""
    spec_type: str = "mel"
    """Type of spectrogram to extract. Can be either 'mel' or 'linear'."""
    fmax: int = 8000
    """Maximum frequency for spectrogram extraction."""


class KerasAppsExtractor(
    FeatureExtractor,
    AudioClipProcessor,
    fname="keras_apps",
    config=KerasAppsExtractorConfig,
):
    """Feature extractor using Keras applications."""

    config: KerasAppsExtractorConfig

    def __init__(self, config: KerasAppsExtractorConfig) -> None:
        super().__init__(config)

        from ertk.tensorflow.utils import init_gpu_memory_growth

        init_gpu_memory_growth()
        _init_models()

        model_args = MODEL_MAP[config.model]
        size = config.size or model_args.size
        self.size = size
        args = {
            "include_top": False,
            "weights": "imagenet",
            "pooling": "avg",
            "input_shape": (size, size, 3),
        }
        self.model = getattr(model_args.module, model_args.funcname)(**args)
        self._preprocess_input = getattr(model_args.module, "preprocess_input")
        _test = self._spectrogram(np.zeros(16000, dtype=np.float32), 16000)
        _test = self._to_img(_test)
        _test = self._preprocess_input(_test)
        _test = self.model(_test[None, :, :, :]).numpy()
        self._dim = _test.shape[-1]

    def _spectrogram(self, x: np.ndarray, sr: float) -> np.ndarray:
        return spectrogram(
            x,
            sr=sr,
            kind=self.config.spec_type,
            pre_emphasis=0,
            window_size=0.025,
            window_shift=0.01,
            n_mels=self.config.n_mels,
            fmin=0,
            fmax=None,
            power=2,
            to_log="db",
        )

    def _to_img(self, spec: np.ndarray) -> np.ndarray:
        if spec.ndim > 2:
            spec = spec.squeeze(0)
        scale = self.size / np.array(spec.shape)
        spec = zoom(spec, scale)
        spec = np.interp(spec, (spec.min(), spec.max()), (0, 255))
        return np.tile(spec[:, :, None], (1, 1, 3))

    def process_instance(self, x: np.ndarray, **kwargs) -> np.ndarray:
        if is_mono_audio(x):
            x = self._spectrogram(x.squeeze(), kwargs.pop("sr"))
        x = self._to_img(x)
        x = self._preprocess_input(x)
        return self.model(x[None, :, :, :]).numpy()

    def process_batch(
        self, batch: Union[Iterable[np.ndarray], np.ndarray], **kwargs
    ) -> List[np.ndarray]:
        batch = list(batch)
        if any(is_mono_audio(x) for x in batch):
            if any(not is_mono_audio(x) for x in batch):
                raise RuntimeError("Inconsistent input representation")
            sr = kwargs.pop("sr")
            batch = [self._spectrogram(x.squeeze(), sr) for x in batch]
        batch = [self._to_img(x) for x in batch]
        batch = np.stack(batch)
        batch = self._preprocess_input(batch)
        return list(self.model(batch).numpy())

    @property
    def dim(self) -> int:
        return self._dim

    @property
    def is_sequence(self) -> bool:
        return False

    @property
    def feature_names(self) -> List[str]:
        return [f"{self.config.model}_{i}" for i in range(self.dim)]
